-- kEnergy
-- Simple energy ticker addon inspired by GonffBar

local name = 'kEnergy'

local function keys(t, k)
    local nk = next(t, k)
    if nk == nil then
        return
    end
    return nk, keys(t, nk)
end
local function empty() return '' end


local candy = AceLibrary('CandyBar-2.0')
local paint = AceLibrary('PaintChips-2.0')
local surface = AceLibrary('Surface-1.0')
local L = AceLibrary('AceLocale-2.2'):new(name)
local addon = AceLibrary('AceAddon-2.0'):new('AceDB-2.0', 'AceEvent-2.0', 'AceConsole-2.0', 'CandyBar-2.0', 'Metrognome-2.0', 'AceDebug-2.0')
local BS = AceLibrary('Babble-Spell-2.2')
local Aura = AceLibrary('SpecialEvents-Aura-2.0')

function addon:OnInitialize()
    surface:Register('BantoBar', 'Interface\\Addons\\'..name..'\\Textures\\banto')
    surface:Register('Smooth', 'Interface\\Addons\\'..name..'\\Textures\\smooth')
    surface:Register('Perl', 'Interface\\Addons\\'..name..'\\Textures\\perl')
    surface:Register('Glaze', 'Interface\\Addons\\'..name..'\\Textures\\glaze')
    surface:Register('Charcoal', 'Interface\\Addons\\'..name..'\\Textures\\Charcoal')
    surface:Register('Otravi', 'Interface\\Addons\\'..name..'\\Textures\\otravi')
    surface:Register('Striped', 'Interface\\Addons\\'..name..'\\Textures\\striped')
    surface:Register('NotSteel', 'Interface\\Addons\\'..name..'\\Textures\\NotSteel')

    self.options = {
        type = 'group',
        args = {
            lock = {
                order = 1,
                type = 'toggle',
                name = L['Lock'],
                desc = L['Lock the bar.'],
                get  = function() return self.locked end,
                set  = function(state) self:Locked(state) end,
            },
            fill = {
                order = 2,
                type = 'toggle',
                name = L['Fill'],
                desc = L['Set the bar to fill rather than empty.'],
                get  = function() return self.db.profile.reversed end,
                set  = function(v)
                    self.db.profile.reversed = v
                    self:SetCandyBarReversed(name, v)
                end,
            },
            shown = {
                order = 3,
                type = 'text',
                name = L['Shown'],
                desc = L['Select when the bar will be shown.'],
                get = function() return self.db.profile.tickstate end,
                set = function(v) self.db.profile.tickstate = v end,
                validate = {
                    ['always'] = L['Always'],
                    ['stealth'] = L['Stealth + Energy Ticking'],
                    ['energyticking'] = L['Energy Ticking'],
                },
            },
            width = {
                order = 4,
                name  = L['Width'],
                desc  = L['Set the width of the bar.'],
                type  = 'range',
                min   = 25,
                max   = 400,
                get   = function() return self.db.profile.width end,
                set   = function(w)
                    self.db.profile.width = w
                    self.frame:SetWidth(w)
                    self:SetCandyBarWidth(name, w)

                    if not self.locked then
                        candy:Update(name)
                    end
                end,
            },
            height = {
                order = 5,
                type  = 'range',
                name  = L['Height'],
                desc  = L['Set the height of the bar.'],
                min   = 5,
                max   = 100,
                get   = function() return self.db.profile.height end,
                set   = function(h)
                    self.db.profile.height = h
                    self.frame:SetHeight(h)
                    self:SetCandyBarHeight(name, h)

                    if not self.locked then
                        candy:Update(name)
                    end
                end,
            },
            texture = {
                order = 6,
                type = 'text',
                name = L['Texture'],
                desc = L['Set the texture of the bar.'],
                get  = function() return self.db.profile.texture end,
                set  = function(v)
                    self.db.profile.texture = v
                    self:SetCandyBarTexture(name, surface:Fetch(v))
                end,
                validate = surface:List(),
            },
            color = {
                order = 7,
                type = 'text',
                name = L['Color'],
                desc = L['Set the color of the bar.'],
                get  = function() return self.db.profile.color end,
                set  = function(v)
                    self.db.profile.color = v
                    self:SetCandyBarColor(name, v)
                end,
                validate = {keys(paint.vars.hex)},
            },
            fontsize = {
                order = 8,
                type = 'range',
                min  = 8,
                max  = 16,
                name = L['Font Size'],
                desc = L['Set the size of the bar text.'],
                get  = function() return self.db.profile.fontsize end,
                set  = function(v)
                    self.db.profile.fontsize = v
                    self:SetFontSize()
                end,
            },
            bartext = {
                order = 9,
                type = 'text',
                name = L['Bar Text'],
                desc = L['Set the text to be shown on the bar.'],
                get  = function() return self.db.profile.bartext end,
                set  = function(v)
                    self.db.profile.bartext = v
                    self:SetText()
                end,
                validate = {
                    ['none'] = L['None'],
                    ['energy'] = L['Energy'],
                    ['cp'] = L['Combo Points'],
                    ['both'] = L['Both'],
                },
            },
            resetpos = {
                order = -5,
                type = 'execute',
                name = L['Reset position'],
                desc = L['Reset bar position.'],
                func = function()
                    self.frame:ClearAllPoints()
                    self.frame:SetPoint('CENTER', UIParent, 'CENTER', 0, 0)
                end,
            },
            tickoffset = {
                order = -4,
                type = 'text',
                name = L['Modify tick time offset'],
                desc = L['Manually modify the offset for the simulated tick time (not persistant).'],
                usage = L['Input a number'],
                get  = function()
                    return self.tickoffset
                end,
                set  = function(v)
                    self.tickoffset = tonumber(v)
                    self:setRate()
                end,
                validate = function(v) return tonumber(v) end,
            },
            debuglevel = {
                order = -1.5,
                type  = 'range',
                name  = L['Debug level'],
                desc  = L['Set addon debugging level'],
                min   = 1,
                max   = 3,
                get   = function() return self:GetDebugLevel() end,
                set   = function(v) self:SetDebugLevel(v) end,
            },
        }
    }
    self.locked = true

    self.playerclass = select(2, UnitClass('player'))
    self:RegisterDB('kEnergySaved')
    self:RegisterDefaults('profile', {
        fontsize  = 12,
        width     = 128,
        height    = 18,
        texture   = 'BantoBar',
        tickstate = 'energyticking',
        reversed  = false,
        bartext   = 'none',
        color     = self.playerclass:lower(),
    })
    if not self.db.char.initialized then
        self.db.char.initialized = true
        self:SetProfile('class')
    end
    self:RegisterChatCommand({'/kenergy', '/ke'}, self.options)
    self:SetDebugLevel(1)
end

function addon:Locked(state)
    self.locked = state

    if self.locked then
        self.frame:Hide()
        self:StopCandyBar(name)
        self:SetText()
    else
        self:StartCandyBar(name)
        self:PauseCandyBar(name)
        candy:Update(name)

        self.frame:Show()
        self.frame.text:SetText(name)
    end

    self.frame:EnableMouse(not state)
end

function addon:DruidInit()
    self:RegisterEvent('SpecialEvents_UnitBuffGained', function(unit, buff)
        if unit == 'player' and buff == BS['Cat Form'] then
            if not self.maxen then
                self.maxen  = UnitManaMax('player')
                self.energy = UnitMana('player')
            end

            self:StartMetro(name)
        end
    end)
    self:RegisterEvent('SpecialEvents_UnitBuffLost', function(unit, buff)
        if unit == 'player' and buff == BS['Cat Form'] then
            prevtime = nil
            self.energyticking = nil
            self:StopMetro(name)
            self:StopCandyBar(name)
        end
    end)

    self.ticktime = 2.0
    self.tickamount = 20

    if Aura:UnitHasBuff(BS['Cat Form']) then
        self.maxen = UnitManaMax('player')
        self.energy = UnitMana('player')
    end
end

function addon:RogueInit()
    self:RegisterEvent('SpecialEvents_UnitBuffGained', function(unit, buff)
        if unit == 'player' and buff == BS['Adrenaline Rush'] then
            self.tickamount = self.tickamount * 2
        end
    end)
    self:RegisterEvent('SpecialEvents_UnitBuffLost', function(unit, buff)
        if unit == 'player' and buff == BS['Adrenaline Rush'] then
            self.tickamount = self.tickamount / 2
        end
    end)

    self.ticktime = 2.0
    self.tickamount = 20

    if Aura:UnitHasBuff(BS['Adrenaline Rush']) then
        self.tickamount = self.tickamount * 2
    end

    self.maxen = UnitManaMax('player')
    self.energy = UnitMana('player')
end

function addon:OnEnable()
    if self.playerclass == 'DRUID' then
        self:DruidInit()
    elseif self.playerclass == 'ROGUE' then
        self:RogueInit()
    else
        DisableAddOn(name)
        return
    end

    self.cp = 0

    self:RegisterEvent('Surface_SetGlobal', function()
        self:SetCandyBarTexture(name, surface:Fetch(self.db.profile.texture))
    end)

    self:RegisterCandyBar(name, self.ticktime, '', nil, self.db.profile.color)
    self:SetCandyBarHeight(name, self.db.profile.height)
    self:SetCandyBarWidth(name, self.db.profile.width)
    self:SetCandyBarTexture(name, surface:Fetch(self.db.profile.texture))
    self:SetCandyBarTimeFormat(name, empty)
    self:SetCandyBarFontSize(name, self.db.profile.fontsize)
    self:SetCandyBarReversed(name, self.db.profile.reversed)

    self:CreateFrame()
    self:RestorePosition()
    self:EnableEventsFrame()

    self:RegisterMetro(name, self.MetroTick, self.ticktime, self)
end

function addon:OnDisable()
    self:UnregisterCandyBar(name)
    self:UnregisterMetro(name)
    self:DisableEventsFrame()
end


_G[name] = addon
