local name = 'kEnergy'
local Aura = AceLibrary('SpecialEvents-Aura-2.0')
local BS = AceLibrary('Babble-Spell-2.2')
local G = AceLibrary('Gratuity-2.0')
local abs = math.abs

-- Scan spellbook to determine the energy costs of our abilities
-- This portion courtesy Grid
function kEnergy:ScanSpellbook()
    local i = 1
    local sName, sRank, sCost

    repeat
        local sIndex, sName, sRank = i, GetSpellName(i, BOOKTYPE_SPELL)
        if sName then
            G:SetSpell(i, BOOKTYPE_SPELL)
            local sCost = G:FindDeformat(ENERGY_COST, 2, 2, false, true)
            if sCost then
                self.spellcosts[sName..'-'..sRank] = tonumber(sCost)
            end
        end
        i = i + 1
    until not sName
end

function kEnergy:EnergyLog(time, amount, msg, ...)
    if self:IsDebugging() then
        if not self.db.raw.EnergyLog then
            self.db.raw.EnergyLog = {}
        end

        local logmsg
        if type(msg) == 'function' then
            logmsg = msg(...)
        else
            logmsg = ('%s [%s]'):format(amount, msg)

            if amount > 0 then
                logmsg = '+' .. logmsg
            end
        end

        self.db.raw.EnergyLog[time] = logmsg
        self:LevelDebug(2, logmsg)
    end
end

function kEnergy:SetText()
    local bartext = self.db.profile.bartext
    local txt
    if bartext == 'cp' then
        txt = tostring(self.cp)
    elseif bartext == 'energy' then
        txt = tostring(self.energy)
    elseif bartext == 'both' then
        txt = ('%s - %s'):format(tostring(self.cp), tostring(self.energy))
    else
        txt = ''
    end
    self.frame.text:SetText(txt)
end

-- Previous tick time
local prevtime
-- Values for our running average / stddev
local count, sum, sumsq
-- Various private pieces of state:
--  pending - anticipated energy changes
--  spellsent - user has initiated a spellcast, but we havent seen the
--              spellcast event yet
--  energyticking - user's energy is not full, so energy ticks are expected
--  barticking - the candybar is currently ticking, relying on its completion
--               function to restart itself.  when this is false, we rely on
--               the metrognome ticker to start the next cycle of bar ticks
local pending, spellsent, firsttick, energyticking, barticking
local function checktick(self)
    return energyticking or
           (self.db.profile.tickstate == 'always') or
           (self.db.profile.tickstate == 'stealth' and ((self.playerclass == 'ROGUE' and Aura:UnitHasBuff('player', BS['Stealth'])) or (self.playerclass == 'DRUID' and Aura:UnitHasBuff('player', BS['Prowl']))))
end

function kEnergy:EnableEventsFrame()
    prevtime = nil
    count, sum, sumsq = 1, 0, 0
    pending = 0
    self.tickoffset = 0
    barticking = false

    self.spellcosts = {}
    self:ScanSpellbook()
    self:RegisterEvent('LEARNED_SPELL_IN_TAB', "ScanSpellbook")

    self:RegisterEvent('UNIT_MAXENERGY', function(unit)
        if unit == 'player' then
            self.maxen = UnitManaMax(unit)
            self:SetText()
        end
    end)

    self:RegisterEvent('PLAYER_COMBO_POINTS', function()
        self.cp = GetComboPoints()
        self:SetText()
    end)

    -- Track energy usage
    self:RegisterEvent('COMBAT_TEXT_UPDATE', function(msgtype, amount, ...)
        if msgtype == 'ENERGY' then
            amount = tonumber(amount)
            pending = pending + amount
            self:EnergyLog(GetTime(), amount, 'Unknown Ability')
        end
    end)

    self:RegisterEvent('UNIT_SPELLCAST_SENT', function(unit, spell, rank, target)
        local cost = self.spellcosts[spell..'-'..rank]
        if cost then
            spellsent = -cost
        else
            spellsent = nil
        end
    end)

    self:RegisterEvent('UNIT_SPELLCAST_SUCCEEDED', function(unit, spell, rank)
        if unit ~= 'player' then
            return
        end

        if spellsent then
            pending = pending + spellsent
            self:EnergyLog(GetTime(), spellsent, spell)
            spellsent = nil
        end
    end)

    -- Anticipate the approaching energy max due to leveling
    self:RegisterEvent('PLAYER_LEVEL_UP', function()
        pending = self.maxen - self.energy
        self:EnergyLog(GetTime(), pending, 'Leveled')
    end)

    self:RegisterEvent('PLAYER_DEAD', function()
        pending = -self.energy
        prevtime = nil
        energyticking = nil
        self:EnergyLog(GetTime(), pending, 'Died')
    end)

    if self.db.raw.EnergyLog then
        self.db.raw.EnergyLog = nil
    end

    self:SetCandyBarCompletion(name, function()
        self:ScheduleEvent(self.CandyBarComplete, 0, self)
    end)
    -- self:SetCandyBarCompletion(name, self.CandyBarComplete, self)

    -- Use a seperate frame for the energy tick, to ensure that no possible
    -- overhead in the event handling introduces a delay in the tick time
    -- measurement
    if not self.events then
        self.events = CreateFrame('Frame', nil, UIParent)
        self.events:SetScript('OnEvent', function(self, event, ...)
            kEnergy.events[event](kEnergy, ...)
        end)

        local function ticklog(time, prevtime)
            return ('Tick%s'):format(prevtime and (" [%s]"):format(time - prevtime) or "")
        end
        function self.events:UNIT_ENERGY(unit)
            local time = GetTime()
            if unit ~= 'player' then
                return
            end

            local olden = self.energy
            self.energy = UnitMana(unit)
            local diff = self.energy - (olden + pending)
            pending = 0

            self:SetText()

            if abs(diff) <= 1 then
                -- Energy alterations were fully anticipated, not a tick
                return
            end

            local notmaxed = true
            if self.energy == self.maxen then
                energyticking = nil
                notmaxed = nil

                if olden + pending >= self.maxen then
                    -- Pending energy alterations caused us to hit max energy--not a tick
                    self:EnergyLog(time, diff, 'Maxed [not a tick]')
                    return
                end
                self:EnergyLog(time, diff, 'Maxed [tick]')
            elseif abs(diff - self.tickamount) > 1 then
                self:EnergyLog(time, diff, 'Unhandled')
                return
            else
                self:EnergyLog(time, diff, ticklog, time, prevtime)
                energyticking = true
            end

            -- Attempt to keep a running standard deviation, so we can throw away
            -- outliers before they screw up our numbers.
            if prevtime then
                local timediff = time - prevtime - self.ticktime
                local sd   = math.sqrt((sumsq - (sum*sum/count))/(count - 1))
                if sd ~= 0 and
                    abs(timediff - self.tickoffset) > 3*sd then
                    -- outlier, throwing out
                    self:LevelDebug(1, 'Outlier detected in tick times, throwing away timediff of %s.', timediff)
                else
                    sum   = sum + timediff
                    sumsq = sum + timediff^2
                    count = count + 1
                    self.tickoffset  = sum / count
                end
            end

            self:setRate()

            if not barticking then
                -- Must be the start of a new cycle of ticks, since the
                -- bar hasn't been ticking.
                barticking = true
                self:StartCandyBar(name)
                self.frame:Show()
            else
                if tickpending or not checktick(self) then
                    -- The current candybar tick awaits its energy tick.
                    -- It has not yet completed, so we'll just have to reset it.
                    self:SetCandyBarTimeLeft(name, 0)
                else
                    -- The candybar restarted itself before receiving its expected
                    -- tick, so let's just adjust its end time to the next
                    -- expected tick.
                    self:SetCandyBarTimeLeft(name, self.ticktime + self.tickoffset)
                end
            end
            tickpending = true

            -- If we maxed our energy, reset the tick cycle via setting prevtime to nil/false
            prevtime = notmaxed and time
        end
    end
    self.events:RegisterEvent('UNIT_ENERGY')
end

function kEnergy:CandyBarComplete()
    -- If this method is called and energy is ticking, the candybar was not
    -- prematurely ended by an energy tick.  We'll be getting a tick soon that
    -- will adjust our end time to appropriate point.

    if checktick(self) then
        self.frame:Show()
        tickpending = not energyticking
        barticking = true
        self:StartCandyBar(name)
    else
        self.frame:Hide()
        tickpending = true
        barticking = false
        self:StartMetro(name)
    end
end

function kEnergy:setRate()
    local totaltime = self.ticktime + self.tickoffset
    self:SetCandyBarTime(name, totaltime)
    self:ChangeMetroRate(name, totaltime)
end

function kEnergy:DisableEventsFrame()
    self.events:UnregisterEvent('UNIT_ENERGY')
end

-- The metrognome timer exists only to have a correct starting point for
-- candybar timers, to ensure they always match up with our energy ticks.
function kEnergy:MetroTick()
    if self.locked and not barticking then
        self:CandyBarComplete()
    end
end
